# Code for the GUI interface to RoutePlanner
#
# Copyright (C) 1996-2001 Chris Lawrence
# This file may be freely distributed under the terms of the RoutePlanner
# license.  A copy should appear as 'LICENSE' in the archive that this
# file was included in.

import gnome, rpdbase

gnome.app_id = 'RoutePlanner'
gnome.app_version = rpdbase.VERID

import string, sys, os, glob, commands, time, libglade, GDK
from rpcitylist import CityList
from rpprogress import ProgressWin

from gtk import *
from gnome.ui import *
from GDK import *
from rpunits import *
import gnome.config

PANES = ['shortroute', 'quickroute', 'prefroute', 'leastfuel']
GUIFILE = '/usr/share/routeplanner/routeplanner.glade'

CONFIG_UNITS   = 'routeplanner/Defaults/units'
CONFIG_DEFMPG  = 'routeplanner/Defaults/default_efficiency'
CONFIG_DEFPREF = 'routeplanner/Defaults/default_preference'

CONFIG_TOLL   = 'routeplanner/Defaults/toll_penalty'
CONFIG_FERRY  = 'routeplanner/Defaults/ferry_penalty'
CONFIG_SCENIC = 'routeplanner/Defaults/scenic_bonus'

class Application:
    def __init__(self, filename=''):
        if os.path.exists('routeplanner.glade'):
            self.wtree = libglade.GladeXML('routeplanner.glade')
        else:
            self.wtree = libglade.GladeXML(GUIFILE)
        
        self.appwin = self.wtree.get_widget('app1')
        self.appwin.connect("destroy", mainquit) 
##        print "Instance self.wtree",self.wtree
##        print "\n***Dictionary ot the instance were I can see self.wtree"
##        print(self.__dict__)
##        print "\nDictionary of the class were I can see on_about_cb"
##        print(Application.__dict__)
        # You need to create this diccionary in order to get the 
        # the bound methods that you can autoconnect with.
        # And quoting James: 
        # "I know it is not perfect, but it should do what you want."
        dict = {}
        for key in dir(self.__class__):
            dict[key] = getattr(self, key)

        self.wtree.signal_autoconnect(dict)
        self.db = None
        self.citylist = self.wtree.get_widget('citylist')
        self.citywin = None
        self.make_widgets_sensitive()
        # Hardcoded for 0.1
        self.units = gnome.config.get_string(CONFIG_UNITS)
        if not self.units or self.units not in VALID_UNITS:
            self.units = UNITS_US
        
        self.currentunits = self.units
        info = {'metric' : UNITS_METRIC, 'us' : UNITS_US,
                'imperial' : UNITS_IMPERIAL}
        for widget, units in info.items():
            if units == self.units:
                self.wtree.get_widget(widget).set_active(TRUE)

        if self.units == UNITS_METRIC:
            self.wtree.get_widget('effunits').set_text('l/100km')

        self.defmpg = gnome.config.get_int(CONFIG_DEFMPG)
        if not self.defmpg or self.defmpg <= 0 or self.defmpg > 100:
            self.defmpg = 20
        effwidget = self.wtree.get_widget('defefficiency')
        effwidget.set_value(self.defmpg)

        self.defpref = gnome.config.get_int(CONFIG_DEFPREF)
        if not self.defpref or self.defpref < 0 or self.defpref > 100:
            self.defpref = 50
        prefwidget = self.wtree.get_widget('defpref')
        prefwidget.set_value(self.defpref)

        self.tollpen = gnome.config.get_int(CONFIG_TOLL)
        if not self.tollpen or self.tollpen < -100 or self.tollpen > 100:
            self.tollpen = 0
        self.wtree.get_widget('avoidtolls').set_active(self.tollpen)
        self.wtree.get_widget('tollpen').set_sensitive(self.tollpen)
        self.wtree.get_widget('tollpen').set_value(self.tollpen)

        self.ferrypen = gnome.config.get_int(CONFIG_FERRY)
        if not self.ferrypen or self.ferrypen < -100 or self.ferrypen > 100:
            self.ferrypen = 0
        self.wtree.get_widget('avoidferries').set_active(self.ferrypen)
        self.wtree.get_widget('ferrypen').set_sensitive(self.ferrypen)
        self.wtree.get_widget('ferrypen').set_value(self.ferrypen)

        self.scenicbonus = gnome.config.get_int(CONFIG_SCENIC)
        if not self.scenicbonus or self.scenicbonus < -100 or \
           self.scenicbonus > 100:
            self.scenicbonus = 0
        self.wtree.get_widget('usescenic').set_active(self.scenicbonus)
        self.wtree.get_widget('scenicbonus').set_sensitive(self.scenicbonus)
        self.wtree.get_widget('scenicbonus').set_value(self.scenicbonus)

        if filename:
            self.open_database(filename)

        prefswin = self.wtree.get_widget('prefswin')
        prefswin.hide_on_delete(TRUE)

        classes = rpdbase.classifications.keys()
        classes.sort()
        classtable = self.wtree.get_widget('classtable')
        classtable.resize(len(classes)-2, 4)

        maxclass = max(rpdbase.classifications.values())+1
        self.speeds = [0]*maxclass
        self.mpgs = [0]*maxclass
        self.prefer = [0]*maxclass

        i = 1
        self.prefwidgets = {}
        for cl in classes:
            classid = rpdbase.classifications[cl]
            if len(rpdbase.defspeed) > classid:
                speed, mpg, pref = (rpdbase.defspeed[classid],
                                    rpdbase.defeff[classid],
                                    rpdbase.defpref[classid])
            else:
                speed, mpg, pref = 50, self.defmpg, self.defpref

            if cl[0:2] == 'XX': continue
            x = GtkLabel(cl)
            x.show()
            classtable.attach(x, 0, 1, i, i+1)
            row = []
            for j in range(1, 4):
                type = ['','speed','efficiency', 'preference'][j]
                x = GtkSpinButton()
                cl2 = string.replace(cl, '/', '_')
                val = gnome.config.get_int('routeplanner/'+type+'/'+cl2)
                maxval = 100
                if not val:
                    if j == 1:
                        val = Distance(speed, UNITS_US).AsUnit(self.units)
                        maxval = 150
                    elif j == 2:
                        val = ConvertEfficiency(mpg, UNITS_US, self.units)
                        
                    else:
                        val = pref
                        
                x.set_digits(0)
                x.get_adjustment().set_all(val, 1, maxval, 1, 10, 1)
                x.set_value(val)
                x.connect('changed', self.prefswin_changed)
                x.show()
                classtable.attach(x, j, j+1, i, i+1)
                row.append(x)
                if j == 1:
                    self.speeds[classid] = val
                elif j == 2:
                    self.mpgs[classid] = val
                else:
                    self.prefer[classid] = val

            self.prefwidgets[cl] = row
            i = i+1

        #print classtable.children()
        classtable.show()
        #prefswin.show()

    def on_exit1_activate(self,*args):
        mainquit()

    def on_about1_activate(self,*args):
        about = self.wtree.get_widget('aboutbox')
        about.close_hides(TRUE)
        about.show()

    def on_about1_close(self,*args):
        self.wtree.get_widget('aboutbox').hide()

    def on_properties1_activate(self, *args):
        if self.db:
            message = 'Current database: %s\n%s\n'\
                      '%d Cites - %d Routes\nAuthor: %s <%s>' % \
                      (self.db.filename, self.db.comment,
                       len(self.db.cities), len(self.db.routes),
                       self.db.author, self.db.author_email)
        else:
            message = 'No database loaded.'
            
        dialog = GnomeOkDialog(message, self.appwin)
        dialog.run_and_close()

    def on_addcity_clicked(self, *args):
        if not self.citywin: self.citywin = CityList(self.wtree)
        self.citywin.startup('', self.db, self.appendcity)

    def on_citylist_button_press_event(self, button, event, *args):
        if event.type == GDK._2BUTTON_PRESS and event.button == 1 \
           and self.citylist.selection:
            self.on_changecity_clicked()

    def on_changecity_clicked(self, *args):
        if not self.citywin:
            self.citywin = CityList(self.wtree)
        pos = self.citylist.selection[0]
        cityname = self.citylist.get_text(pos, 0)
        self.citywin.startup(cityname, self.db, self.changecity, pos)

    def appendcity(self, city, ind, *data):
        self.citylist.append([city])
        self.citylist.set_row_data(self.citylist.rows-1, ind)
        self.citylist.select_row(self.citylist.rows-1, 0)
        self.make_widgets_sensitive()

    def changecity(self, city, ind, pos):
        self.citylist.freeze()
        self.citylist.insert(pos, [city])
        self.citylist.set_row_data(pos, ind)
        self.citylist.remove(pos+1)
        self.citylist.select_row(pos, 0)
        self.citylist.thaw()

    def on_removecity_clicked(self, *args):
        pos = self.citylist.selection[0]
        self.citylist.remove(pos)
        rows = self.citylist.rows
        pos = min(rows-1, pos)
        self.citylist.select_row(pos, 0)
        self.make_widgets_sensitive()

    def on_up_clicked(self, *args):
        pos = self.citylist.selection[0]
        self.citylist.swap_rows(pos, pos-1)
        self.citylist.select_row(pos-1, 0)
        self.update_ud_widgets()

    def on_down_clicked(self, *args):
        pos = self.citylist.selection[0]
        self.citylist.swap_rows(pos, pos+1)
        self.citylist.select_row(pos+1, 0)
        self.update_ud_widgets()

    def update_ud_widgets(self):
        wlist = {'up': FALSE, 'down': FALSE,
                 'changecity' : TRUE, 'removecity' : TRUE}
        row = self.citylist.selection[0]
        rows = self.citylist.rows
        if row > 0:
            wlist['up'] = TRUE
        if row < rows-1:
            wlist['down'] = TRUE
        
        for widget, state in wlist.items():
            self.wtree.get_widget(widget).set_sensitive(state)

    def on_citylist_select_row(self, *args):
        self.update_ud_widgets()

    def on_citylist_unselect_row(self, *args):
        wlist = ['up', 'down', 'changecity', 'removecity']
        for widget in wlist:
            self.wtree.get_widget(widget).set_sensitive(FALSE)

    def on_filesel_ok_clicked(self, *args):
        filename = self.fw.get_filename()
        self.fw.destroy()
        self.open_database(filename)

    def on_open1_activate(self, *args):
        self.fw = GtkFileSelection('Select database')
        self.fw.set_transient_for(self.appwin)
        self.fw.selection_entry.connect("activate",
                                         self.on_filesel_ok_clicked)
        self.fw.ok_button.connect("pressed", self.on_filesel_ok_clicked)
        self.fw.cancel_button.connect("pressed", self.fw.destroy)
        self.fw.set_modal(TRUE)
        self.fw.hide_on_delete(TRUE)
        self.fw.set_filename('/usr/share/routeplanner/')
        self.fw.show()

    def on_plan_clicked(self, *args):
        db = self.db
        path = []
        for i in range(self.citylist.rows):
            cname = self.citylist.get_text(i, 0)
            cpos = self.citylist.get_row_data(i)
            info = db.cities[cpos]
            if path and info == path[-1]:
                dialog = GnomeErrorDialog('Same city appears twice in a '
                                          'row:\n'+cname+'\n'
                                          'Please fix this and plan again.',
                                          self.appwin)
                dialog.run_and_close()
                return
            path.append(info)

        methods = len(rpdbase.methods)
        displist = [None]*methods
        self.routes = [None]*methods
        progwin = ProgressWin("Calculating route", self.appwin)
        progwin.set_label('Planning')
        progwin.set_current(0)
        progwin.set_max(methods)
        progwin.show()

        for method in range(methods):
            progwin.set_label('Planning '+rpdbase.methods[method])
            progwin.set_current(method)
            trail = []
            for i in range(len(path)-1):
                trail.append( db.NewNavigate(
                    path[i], path[i+1], method = method,
                    codetable = self.speeds, prefertable = self.prefer,
                    defaultprefer = self.defpref, mpgtable = self.mpgs,
                    defmpg = self.defmpg,
                    flagbonus = { 'toll' : -self.tollpen,
                                  'scenic' : self.scenicbonus,
                                  'ferry': -self.ferrypen } ))
            progwin.set_current(method+0.5)

            displist[method] = rpdbase.ProduceTrail(trail, db, self.units,
                                                    self.mpgs, self.defmpg,
                                                    self.speeds)

            del trail # Free some memory
            progwin.set_current(method+1)
        
        progwin.destroy()

        planwin = self.wtree.get_widget('planwin')
        startloc = self.wtree.get_widget('startloc')
        endloc = self.wtree.get_widget('endloc')
        vialocs = self.wtree.get_widget('vialocs')

        start, end, via = path[0], path[-1], path[1:-1]
        startloc.set_text(str(start))
        endloc.set_text(str(end))
        if via:
            vialocs.set_text(string.join(map(str, via), '; '))
        else:
            vialocs.set_text('[none]')

        font = load_font("fixed")
        black = GdkColor(0,0,0)

        widgets = PANES
        for j in range(methods):
            route = self.wtree.get_widget(widgets[j])
            conlen = route.get_length()
            route.delete_text(0, conlen)
            text = rpdbase.FormatRoute(displist[j], path, self.units)
            route.insert(font, black, None, text)
            self.routes[j] = text
        
        planwin.connect("delete_event", self.close_route_win)
        planwin.show()

    def on_copy1_activate(self, *args):
        pane = self.wtree.get_widget('notebook1').get_current_page()
        widget = PANES[pane]
        self.wtree.get_widget(widget).copy_clipboard()

    def on_close1_activate(self, *args):
        self.close_route_win()

    def close_route_win(self, *args):
        self.wtree.get_widget('planwin').hide()
        return TRUE

    def on_print1_activate(self, *args):
        printwin = self.wtree.get_widget('printwin')
        printwin.connect("delete_event", self.close_print_win)
        printwin.set_transient_for(self.wtree.get_widget('planwin'))
        printwin.hide_on_delete(TRUE)
        printwin.show()

    def close_print_win(self, *args):
        self.wtree.get_widget('printwin').hide()
        return TRUE

    def on_print_ok_clicked(self, *args):
        self.wtree.get_widget('printwin').hide()
        command = self.wtree.get_widget('printcomm').get_text()
        items = []
        items.append( self.wtree.get_widget('shortest').get_active() )
        items.append( self.wtree.get_widget('quickest').get_active() )
        items.append( self.wtree.get_widget('preferred').get_active() )
        items.append( self.wtree.get_widget('lowestfuel').get_active() )

        separator = '\n\n'
        if self.wtree.get_widget('use_ff').get_active():
            separator = '\f'

        fh = os.popen(command, 'w')
        already_printed = 0

        for j in range(len(rpdbase.methods)):
            if items[j]:
                if already_printed: fh.write(separator)
                fh.write(rpdbase.methods[j]+" Route:\n\n")
                fh.write(self.routes[j]+'\n')
                already_printed = 1

        fh.close()

        self.wtree.get_widget('printwin').hide()
        return TRUE

    def on_save_as1_activate(self, *args):
        self.fw = GtkFileSelection('Save routes as...')
        self.fw.set_transient_for(self.wtree.get_widget('planwin'))
        self.fw.selection_entry.connect("activate", self.save_as_done)
        self.fw.ok_button.connect("pressed", self.save_as_done)
        self.fw.cancel_button.connect("pressed", self.fw.destroy)
        self.fw.set_modal(TRUE)
        self.fw.hide_on_delete(TRUE)
        self.fw.show()

    def save_as_done(self, *args):
        fn = self.fw.get_filename()
        if os.path.exists(fn):
            dialog = GnomeMessageBox("Overwrite existing file\n"+fn+"?",
                                     MESSAGE_BOX_QUESTION, STOCK_BUTTON_YES,
                                     STOCK_BUTTON_NO)
            dialog.set_parent(self.wtree.get_widget('planwin'))
            if dialog.run_and_close():
                return

        self.fw.destroy()
        del self.fw
        try:
            fh = open(fn, 'w')
        except IOError, x:
            dialog = GnomeErrorDialog(str(x), self.wtree.get_widget('planwin'))
            dialog.run_and_close()
            return
        
        for j in range(len(rpdbase.methods)):
            if j: fh.write('\n\n')
            fh.write(rpdbase.methods[j]+" Route:\n\n")
            fh.write(self.routes[j]+'\n')
        fh.close()

    def make_widgets_sensitive(self):
        if not self.db:
            wlist = {'addcity' : FALSE,
                     'citylist' : FALSE,
                     'plan' : FALSE}
        elif self.citylist.rows > 1:
            wlist = {'addcity' : TRUE,
                     'citylist' : TRUE,
                     'plan' : TRUE}
        else:
            wlist = {'addcity' : TRUE,
                     'citylist' : TRUE,
                     'plan' : FALSE}

        for widget, state in wlist.items():
            self.wtree.get_widget(widget).set_sensitive(state)

    def open_database(self, filename):
        progwin = ProgressWin(os.path.basename(filename)+'...', self.appwin)
        progwin.show()
        try:
            self.db = rpdbase.RPDatabase(filename, quiet=1,
                                         progressbar=progwin)
        except (rpdbase.invalidformat, IOError), x:
            progwin.destroy()
            error = GnomeErrorDialog('Unable to open '+filename+':\n\n'+
                                     str(x), self.appwin)
            error.run_and_close()
            return

        self.make_widgets_sensitive()
        self.citylist.clear()
        self.appwin.set_title('RoutePlanner: '+filename)
        progwin.destroy()

    def on_preferences1_activate(self, *args):
        prefswin = self.wtree.get_widget('prefswin')
        prefswin.close_hides(TRUE)
        prefswin.show()

    def units_changed(self, *args):
        self.prefswin_changed()
        oldunits = newunits = self.currentunits
        info = {'metric' : UNITS_METRIC, 'us' : UNITS_US,
                'imperial' : UNITS_IMPERIAL}
        for widget, units in info.items():
            if self.wtree.get_widget(widget).get_active():
                newunits = units

        if oldunits == newunits:
            return

        if newunits == UNITS_METRIC:
            self.wtree.get_widget('effunits').set_text('l/100km')
        else:
            self.wtree.get_widget('effunits').set_text('mpg')

        effwidget = self.wtree.get_widget('defefficiency')
        current = effwidget.get_value()
        current = ConvertEfficiency(current, oldunits, newunits)
        effwidget.set_value(current)

        for cl, row in self.prefwidgets.items():
            for j in range(1, 3):
                val = row[j-1].get_value()
                if j == 1:
                    row[j-1].set_value(Distance(val, oldunits).
                                       AsUnit(newunits))
                else:
                    row[j-1].set_value(ConvertEfficiency(val, oldunits,
                                                       newunits))
        
        self.currentunits = newunits

    def prefswin_changed(self, *args):
        self.wtree.get_widget('prefswin').set_modified(TRUE)

    def on_prefswin_apply(self, *args):
        # Update prefs here
        info = {'metric' : UNITS_METRIC, 'us' : UNITS_US,
                'imperial' : UNITS_IMPERIAL}
        for widget, units in info.items():
            if self.wtree.get_widget(widget).get_active():
                self.units = units

        self.defmpg = self.wtree.get_widget('defefficiency').get_value_as_int()
        self.defpref = self.wtree.get_widget('defpref').get_value_as_int()

        gnome.config.set_string(CONFIG_UNITS, self.units)
        gnome.config.set_int(CONFIG_DEFMPG, self.defmpg)
        gnome.config.set_int(CONFIG_DEFPREF, self.defpref)

        self.tollpen = (self.wtree.get_widget('avoidtolls').get_active() and
                        self.wtree.get_widget('tollpen').get_value_as_int())
        self.ferrypen = (self.wtree.get_widget('avoidferries').get_active()
                         and self.wtree.get_widget('ferrypen').
                         get_value_as_int())
        self.scenicbonus = (self.wtree.get_widget('usescenic').get_active()
                            and self.wtree.get_widget('scenicbonus').
                            get_value_as_int())
        gnome.config.set_int(CONFIG_TOLL, self.tollpen)
        gnome.config.set_int(CONFIG_FERRY, self.ferrypen)
        gnome.config.set_int(CONFIG_SCENIC, self.scenicbonus)
        
        items = self.prefwidgets.items()
        items.sort()
        for cl, row in items:
            classid = rpdbase.classifications[cl]
            for j in range(1, 4):
                type = ['','speed','efficiency', 'preference'][j]
                val = row[j-1].get_value()

                cl = string.replace(cl, '/', '_')
                gnome.config.set_int('routeplanner/'+type+'/'+cl, val)

                if j == 1:
                    self.speeds[classid] = val
                elif j == 2:
                    self.mpgs[classid] = val
                else:
                    self.prefer[classid] = val
                
        gnome.config.sync()

    def on_avoidtolls_toggled(self, widget, *args):
        self.wtree.get_widget('tollpen').set_sensitive(widget.get_active())
        self.prefswin_changed()

    def on_avoidferries_toggled(self, widget, *args):
        self.wtree.get_widget('ferrypen').set_sensitive(widget.get_active())
        self.prefswin_changed()

    def on_usescenic_toggled(self, widget, *args):
        self.wtree.get_widget('scenicbonus').set_sensitive(widget.get_active())
        self.prefswin_changed()

    def on_prefswin_close(self, *args):
        self.wtree.get_widget('prefswin').hide()
        return TRUE

def main(filename=''):
    app = Application(filename)
    mainloop()
